import logging
import numpy as np
from dateutil import tz
import ast

import ujson

import sys
import datetime as dt
from decimal import Decimal

import io

try:
    import pandas as pd
except:
    logging.info('Pandas is not installed, Cursor.select_as_dataframe is not available.')

class NestingLevelTooHigh(Exception):
    pass



class DictionaryAdapter(object):
    def getfields(self, dict):
        return dict.keys()

    def getval(self, dict, field):
        if field in dict:
            return dict[field]
        else:
            return None

class MultilevelDictionaryAdapter(object):
    def getfields(self, doc, prefix='', had_array=False):
        result = []
        for k, v in doc.items():
            if isinstance(v, dict):
                result.extend(self.getfields(v, prefix + k + '_', had_array))
            elif hasattr(v, '__iter__') and not isinstance(v, str):
                if had_array:
                    raise NestingLevelTooHigh()
                dict_keys = set()
                for tmp in v:
                    if isinstance(tmp, dict):
                        try:
                            subkeys = self.getfields(tmp, prefix + k + '_', True)
                            dict_keys = dict_keys.union(set(subkeys))
                        except NestingLevelTooHigh:
                            result.append(prefix + k + '_json')
                            break
                    else:
                        result.append(prefix + k)
                        break
                result.extend(dict_keys)
            else:
                result.append(prefix+k)
        return result

    def getval(self, doc, field):
        return self._getval_rec(doc, field.split('.'))

    def _getval_rec(self, val, parts):
        if len(parts) == 0:
            return val
        part = parts[0]
        if len(parts) == 1 and part == 'json':
            return ujson.dumps(val)
        if isinstance(val, dict):
            if part not in val:
                return None
            return self._getval_rec(val[part], parts[1:])
        else:
            assert hasattr(val, '__iter__')
            return [self._getval_rec(x[part], parts[1:]) if part in x else None for x in val]

class ObjectAdapter(object):
    def getfields(self, obj):
        return [x for x in dir(obj) if not x.startswith('__')]

    def getval(self, obj, field):
        return getattr(obj, field)

class TabSeparatedWithNamesAndTypesFormatter(object):
    def __init__(self):
        self.enable_map_datatype = False

    def generalize_type(self, existing_type, new_type):
        arr = 'Array('
        nu = 'Nullable('
        ma = 'Map('
        if existing_type == new_type:
            return existing_type
        elif existing_type.startswith(arr) and new_type.startswith(arr):
            return 'Array(%s)' % self.generalize_type(existing_type[len(arr):-1], new_type[len(arr):-1])
        elif existing_type.startswith(ma) and new_type.startswith(ma):
            existing_spec = existing_type[len(ma):-1].split(',')
            new_spec = new_type[len(ma): -1].split(',')
            return 'Map(%s,%s)' % (self.generalize_type(existing_spec[0].strip(), new_spec[0].strip()), self.generalize_type(
                existing_spec[1].strip(), new_spec[1].strip()))
        elif existing_type.startswith(arr) or new_type.startswith(arr):
            raise Exception('Cannot generalize %s and %s' % (existing_type, new_type))
        elif existing_type.startswith(ma) or new_type.startswith(ma):
            raise Exception('Cannot generalize %s and %s' % (existing_type, new_type))
        elif existing_type.startswith(nu) or new_type.startswith(nu):
            if existing_type.startswith(nu):
                existing_type = existing_type[len(nu):-1]
            if new_type.startswith(nu):
                new_type = new_type[len(nu):-1]
            return 'Nullable(%s)' % self.generalize_type(existing_type, new_type)
        elif (existing_type.startswith('Int') or existing_type.startswith('UInt')) and new_type.startswith('Float'):
            return new_type
        elif existing_type.startswith('Float') and (new_type.startswith('Int') or new_type.startswith('UInt')):
            return existing_type
        elif existing_type.startswith('Int') and new_type.startswith('Int'):
            existing_bits = int(existing_type[3:])
            new_bits = int(new_type[3:])
            return 'Int%d' % (max(existing_bits, new_bits))
        elif existing_type.startswith('UInt') and new_type.startswith('UInt'):
            existing_bits = int(existing_type[4:])
            new_bits = int(new_type[4:])
            return 'UInt%d' % (max(existing_bits, new_bits))
        elif existing_type.startswith('Float') and new_type.startswith('Float'):
            existing_bits = int(existing_type[5:])
            new_bits = int(new_type[5:])
            return 'Float%d' % (max(existing_bits, new_bits))
        elif existing_type == 'Date' and new_type.startswith('DateTime'):
            return new_type
        elif existing_type.startswith('DateTime') and new_type == 'Date':
            return existing_type
        elif existing_type.startswith('DateTime(') and new_type == 'DateTime':
            return existing_type
        elif existing_type == 'DateTime' and new_type.startswith('DateTime('):
            return new_type
        return 'String'

    def is_compatible_type(self, existing_type, new_type):
        arr = 'Array('
        nu = 'Nullable('
        ma = 'Map('
        if existing_type == new_type:
            return True
        elif existing_type.startswith(arr) and new_type.startswith(arr):
            return self.is_compatible_type(existing_type[len(arr):-1], new_type[len(arr):-1])
        elif existing_type.startswith(ma) and new_type.startswith(ma):
            existing_spec = existing_type[len(ma):-1].split(',')
            new_spec = new_type[len(ma): -1].split(',')
            return self.is_compatible_type(existing_spec[0].strip(), new_spec[0].strip()) \
                   and self.is_compatible_type(existing_spec[1].strip(),
                                              new_spec[1].strip())
        elif existing_type.startswith(arr) or new_type.startswith(arr):
            return False
        elif existing_type.startswith(ma) or new_type.startswith(ma):
            return False
        elif existing_type.startswith(nu) or new_type.startswith(nu):
            if existing_type.startswith(nu):
                existing_type = existing_type[len(nu):-1]
            if new_type.startswith(nu):
                new_type = new_type[len(nu):-1]
            return self.is_compatible_type(existing_type, new_type)
        elif (existing_type.startswith('Int') or existing_type.startswith('UInt')) and new_type.startswith('Float'):
            return False
        elif existing_type.startswith('Float') and (new_type.startswith('Int') or new_type.startswith('UInt')):
            return True
        elif existing_type.startswith('Int') and new_type.startswith('Int'):
            existing_bits = int(existing_type[3:])
            new_bits = int(new_type[3:])
            return new_bits <= existing_bits
        elif existing_type.startswith('UInt') and new_type.startswith('UInt'):
            existing_bits = int(existing_type[4:])
            new_bits = int(new_type[4:])
            return new_bits <= existing_bits
        elif existing_type.startswith('Float') and new_type.startswith('Float'):
            existing_bits = int(existing_type[5:])
            new_bits = int(new_type[5:])
            return new_bits <= existing_bits
        elif existing_type == 'Date' and new_type.startswith('DateTime'):
            return False
        elif existing_type.startswith('DateTime') and new_type == 'Date':
            return False
        elif existing_type.startswith('DateTime(') and new_type == 'DateTime':
            return True
        elif existing_type == 'DateTime' and new_type.startswith('DateTime('):
            return True
        return False


    def clickhousetypefrompython(self, pythonobj, name, nullablelambda=lambda fieldname: False):
        if pythonobj is None:
            raise Exception('Cannot infer type of "%s" from None' % name)
        result = None
        try:
            isstring = isinstance(pythonobj, basestring)
            islong = isinstance(pythonobj, long)
        except:
            isstring = isinstance(pythonobj, str)
            islong = isinstance(pythonobj, int)
        if isstring:
            result = 'String'
        elif isinstance(pythonobj, str):
            result = 'String'
        elif isinstance(pythonobj, bool):
            result = 'UInt8'
        elif isinstance(pythonobj, int) or islong:
            result = 'Int64'
        elif isinstance(pythonobj, float) or isinstance(pythonobj, Decimal):
            result = 'Float64'
        elif isinstance(pythonobj, dt.datetime):
            if pythonobj.tzname() is not None:
                result = "DateTime('%s')" % pythonobj.tzname()
            else:
                result = 'DateTime'
        elif isinstance(pythonobj, dt.date):
            result = 'Date'
        elif isinstance(pythonobj, dict):
            if self.enable_map_datatype:
                return 'Map(%s, %s)' % (self.clickhousetypefrompython(list(pythonobj.keys())[0],'', nullablelambda),
                                       self.clickhousetypefrompython(list(pythonobj.values())[0], '', nullablelambda),
                                       )
            else:
                result = 'String' # Actually JSON
        elif hasattr(pythonobj, '__iter__') and not isinstance(pythonobj, str):
            possibletypes = set()
            for x in pythonobj:
                if x is not None:
                    possibletypes.add(self.clickhousetypefrompython(x, name, nullablelambda))
            if len(possibletypes) == 1:
                result = 'Array(' + list(possibletypes)[0]  + ')'
            elif len(possibletypes) == 0:
                raise Exception('Cannot infer type of "%s" from empty array' % name)
            else:
                possibletypes = list(possibletypes)
                type = possibletypes[0]
                for other in possibletypes[1:]:
                    type = self.generalize_type(type, other)
                result = 'Array(' + type  + ')'
        if result is None:
            raise Exception('Cannot infer type of "%s", type not supported for: %s, %s' % (name, repr(pythonobj), type(pythonobj)))
        if nullablelambda(name) and not result.startswith('Array'):
            result = 'Nullable(%s)' % result
        return result


    def get_schema(self, doc, nullablelambda=lambda fieldname: False):
        if isinstance(doc, dict):
            adapter = DictionaryAdapter()
        else:
            adapter = ObjectAdapter()

        fields = adapter.getfields(doc)
        types = [self.clickhousetypefrompython(adapter.getval(doc, f), f, nullablelambda) for f in fields]

        return fields, types

    def format(self, rows, fields=None, types=None):
        if len(rows) == 0:
            raise Exception('No data in rows')

        if fields is None and types is None:
            fields, types = self.get_schema(rows[0])

        if sys.version_info[0] == 2:
            fields = [x.encode('utf8') for x in fields]

        if isinstance(rows[0], dict):
            adapter = DictionaryAdapter()
        else:
            adapter = ObjectAdapter()

        return fields, types, '%s\n%s\n%s' % (
            '\t'.join(fields),
            '\t'.join(types),
            '\n'.join(['\t'.join([self.formatfield(adapter.getval(r, f), t, f) for f, t in zip(fields, types)]) for r in rows])
        )

    def formatfield(self, value, type, name, inarray = False):
        try:
            if type.startswith('LowCardinality(') and type.endswith(')'):
                type = type[len('LowCardinality('):-1]

            if type.startswith('Nullable(') and type.endswith(')'):
                type = type[len('Nullable('):-1]
                if value is None:
                    return '\\N'

            if type in ['UInt8','UInt16', 'UInt32', 'UInt64','Int8','Int16','Int32','Int64']:
                if value is None:
                    return '0'
                if isinstance(value, bool):
                    return '1' if value else '0'
                return str(value)
            if type in ['String', 'IPv6']:
                if value is None:
                    escaped = ''
                else:
                    # Because String is also a type for columns having values of various type depending on row, the value
                    # parameter might be just anything and has to be converted first
                    if sys.version_info[0] == 2 and isinstance(value, unicode):
                        value = value.encode('utf8')
                    elif sys.version_info[0] == 3 and isinstance(value, bytes):
                        value = value.decode('utf8')
                    if not isinstance(value, str):
                        if isinstance(value, int) or isinstance(value, float) or isinstance(value, bool):
                            value = str(value)
                        elif isinstance(value, dt.date):
                            value = value.strftime('%Y-%m-%d')
                        elif isinstance(value, dt.datetime):
                            value = value.strftime('%Y-%m-%d %H:%M:%S')
                        else:
                            value = ujson.dumps(value)
                    escaped =  value.replace('\\','\\\\').replace('\n','\\n').replace('\t','\\t')
                if inarray:
                    return "'%s'" % escaped.replace("'", "\\'")
                else:
                    return  escaped
            if type in ['Float32', 'Float64']:
                if value is None:
                    return '0.0'
                return str(value).replace(',','.') # replacing comma to dot to ensure US format
            if type == 'Date':
                if value is None or value <= dt.date(1970,1,2):
                    escaped = '0000-00-00'
                else:
                    escaped = '%04d-%02d-%02d' % (value.year, value.month, value.day)
                if inarray:
                    return "'%s'" % escaped
                else:
                    return escaped
            if type == 'DateTime' or type.startswith('DateTime('):
                if value is None or value.replace(tzinfo=None) <= dt.datetime(1970,1,2,0,0,0):
                    escaped = '0000-00-00 00:00:00'
                else:
                    escaped = '%04d-%02d-%02d %02d:%02d:%02d' % (value.year, value.month, value.day, value.hour, value.minute, value.second)
                if inarray:
                    return "'%s'" % escaped
                else:
                    return escaped
            if 'Array' in type:
                if value is None:
                    return '[]'
                return '[%s]' % ','.join([self.formatfield(x, type[6:-1], name, True) for x in value])
            if type.startswith('Map(') and type.endswith(')'):
                if value is None:
                    return '{}'
                spec=type[4:-1].split(',')
                if len(spec) != 2:
                    raise Exception('Cannot format type %s, it is either malformed or too nested for this simple '
                                    'driver' % (type))
                return '{%s}' % ','.join(['%s:%s' % (
                    self.formatfield(k, spec[0].strip(), name+str(k), True),
                    self.formatfield(v, spec[1].strip(), name+str(v), True)) for k, v in value.items()])
        except Exception as e:
            if sys.version_info[0] == 3:
                raise Exception('Cannot format field %s' % name) from e
            else:
                raise Exception('Cannot format field %s, %s' % (name, e))

        raise Exception('Unexpected error, field %s cannot be formatted, %s, %s' % (name, str(value), type))


    def unformatfield(self, value, type):
        if type.startswith('LowCardinality(') and type.endswith(')'):
            type = type[len('LowCardinality('):-1]

        if type.startswith('Nullable(') and type.endswith(')'):
            type = type[len('Nullable('):-1]
            if value == '\\N':
                return None

        if type in ['UInt8','UInt16', 'UInt32', 'UInt64','Int8','Int16','Int32','Int64']:
            return int(value)
        if type in ['String', 'IPv6', 'UUID']:
            return value.replace('\\n','\n').replace('\\t','\t').replace("\\'", "'").replace('\\\\','\\')
        if type in ['Float32', 'Float64']:
            return float(value)
        if type == 'Date':
            if value.startswith("'"):
                value = value[1:]
            if value.endswith("'"):
                value = value[:-1]
            if value == '0000-00-00' or value == '1970-01-01':
                return None
            return dt.datetime.strptime(value, '%Y-%m-%d').date()
        if type == 'DateTime' or type.startswith('DateTime('):
            if value.startswith("'"):
                value = value[1:]
            if value.endswith("'"):
                value = value[:-1]
            if value == '0000-00-00 00:00:00' or value == '1970-01-01 86:28:16':
                return None
            tmp = dt.datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
            if type.startswith('DateTime('):
                tmp = tmp.replace(tzinfo=tz.gettz(type.split('(')[1][2:-3]))
            return tmp
        if type.startswith('Array('):
            value = value.replace(',,', ",'',") # workaround empty string clickhouse bug
            parts = ast.literal_eval(value)

            def handle_dates(val, typ):
                if typ.startswith('Date'):
                    return [self.unformatfield(x, typ) for x in val]
                else:
                    if typ.startswith('Array('):
                        return [handle_dates(x, typ[6:-1].strip()) for x in val]
                    else:
                        return val

            return handle_dates(parts, type[6:-1].strip())

        if type.startswith('Map(') and type.endswith(')'):
            spec=type[4:-1].split(',')
            if len(spec) != 2:
                raise Exception('Cannot unformat type %s, it is either malformed or too nested for this simple '
                                'driver' % (type))
            x = ast.literal_eval(value)
            spec[0] = spec[0].strip()
            spec[1] = spec[1].strip()
            result = dict()
            for k, v in x.items():
                if spec[0].startswith('Date'):
                    k = self.unformatfield(k, spec[0])
                if spec[1].startswith('Date'):
                    v = self.unformatfield(v, spec[1])
                elif spec[1].startswith('Array(') and spec[1].endswith(')'):
                    v = self.unformatfield(v, spec[1][6:-1])
                result[k] = v
            return result
        raise Exception('Unexpected error, field cannot be unformatted, %s, %s' % (str(value), type))

    def unformat_as_dataframe(self, payload_b):
        if sys.version_info[0] == 3:
            payload = payload_b.decode('utf8')
        else:
            payload = payload_b
        split = payload.split('\n')
        if len(split) < 3:
            raise Exception('Unexpected error, no result')

        fields = split[0].split('\t')
        types = split[1].split('\t')

        dtypes = dict()
        converters = dict()

        def make_to_datetime(type):
            if type.startswith('('):
                tzinfo = tz.gettz(type.split('(')[1][2:-3])
                def to_datetime(value):
                    if value.startswith("'"):
                        value = value[1:]
                    if value.endswith("'"):
                        value = value[:-1]
                    if value == '0000-00-00 00:00:00' or value == '1970-01-01 86:28:16':
                        return None
                    return dt.datetime.strptime(value, '%Y-%m-%d %H:%M:%S').replace(tzinfo=tzinfo)
            else:
                def to_datetime(value):
                    if value.startswith("'"):
                        value = value[1:]
                    if value.endswith("'"):
                        value = value[:-1]
                    if value == '0000-00-00 00:00:00' or value == '1970-01-01 86:28:16':
                        return None
                    return dt.datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
            return to_datetime

        def to_date(value):
            if value.startswith("'"):
                value = value[1:]
            if value.endswith("'"):
                value = value[:-1]
            if value == '0000-00-00' or value == '1970-01-01':
                return None
            return dt.datetime.strptime(value, '%Y-%m-%d').date()

        for field, type in zip(fields, types):
            if type.startswith('LowCardinality(') and type.endswith(')'):
                type = type[len('LowCardinality('):-1]

            if type.startswith('Nullable(') and type.endswith(')'):
                type = type[len('Nullable('):-1]

            if 'Array' in type or 'Map' in type:
                dtypes[field] = np.object
            elif type in ['UInt8', 'UInt16', 'UInt32', 'UInt64', 'Int8', 'Int16', 'Int32', 'Int64']:
                dtypes[field] = np.float # np.int does not support NANs
            elif type in ['String', 'IPv6']:
                dtypes[field] = str
            elif type in ['Float32', 'Float64']:
                dtypes[field] = np.float
            elif type == 'Date':
                converters[field] = to_date
            elif type == 'DateTime' or type.startswith('DateTime('):
                converters[field] = make_to_datetime(type)
            else:
                raise Exception('type %s is not supported' % type)


        df = pd.read_csv(
            io.StringIO(payload),
            sep='\t',
            na_values='\\N',
            dtype = dtypes,
            converters= converters,
            header=0,
            skiprows=[1])

        return df

    def unformat(self, payload_b):
        if sys.version_info[0] == 3:
            payload = payload_b.decode('utf8')
        else:
            payload = payload_b
        payload = payload.split('\n')
        if len(payload) < 3:
            raise Exception('Unexpected error, no result')

        fields = payload[0].split('\t')
        types = payload[1].split('\t')
        result = []
        for line in payload[2:-1]:
            line = line.split('\t')
            d = dict()
            for l, t, f in zip(line, types, fields):
                d[f] = self.unformatfield(l,t)
            result.append(d)

        return result


# Testing
if __name__ == '__main__':

    class DTO:
        def __init__(self):
            self.id = 1
            self.firm = 'ACME, Inc'
            self.budget = 3.1415
            self.paid = True
            self.lastuseddate = dt.datetime.now()
            self.escaping = '"\t\n\''

    data = [DTO(), DTO(), DTO()]

    formatter = TabSeparatedWithNamesAndTypesFormatter()
    print()
    fields, types, f = formatter.format(data)
    print(f)
    v = formatter.unformat(f)
    print(v)

    print(data[0].escaping)
    print(v[0]['escaping'])
    print(data[0].escaping == v[0]['escaping'])

    print (formatter.get_schema({'id': 3, 'Offer': {'price': 5, 'count': 1}, 'Images': [{'file': 'a', 'size': 400, 'tags': ['cool','Nikon']}, {'file': 'b', 'size': 500}]}))

    print (formatter.get_schema({'foo': ['Offer']}))



import urllib.request, urllib.parse, urllib.error
import multiprocessing
import logging
import traceback
import base64
import os

import requests
from requests.adapters import HTTPAdapter

from pyclickhouse.Cursor import Cursor


class Connection(object):
    """
    Represents a Connection to Clickhouse. Because HTTP protocol is used underneath, no real Connection is
    created. The Connection is rather an temporary object to create cursors.

    Clickhouse does not support transactions, thus there is no commit method. Inserts are commited automatically
    if they don't produce errors.
    """
    Session=None
    Pool_connections=1
    Pool_maxsize=10

    def __init__(self, host, port=None, username='default', password='', pool_connections=1, pool_maxsize=10,
                 timeout=5, clickhouse_settings='', auth_method=None, use_own_session=False, secure=None,
                 server_cert=True):
        """
        Create a new Connection object. Because HTTP protocol is used underneath, no real Connection is
        created. The Connection is rather an temporary object to create cursors.

        Before using a cursor, you can optionally call "open" method of Connection. This method will check
        whether Clickhouse is responding and raise an Exception if it isn't. If you get the cursor from this
        Connection, it will be automatically opened for you.

        Note that the Connection may not be reused between multiprocessing-Processes - create an own Connection per Process.

        :param host: hostname or ip of clickhouse host (without http://)
        :param port: port of the Http interface (usually 8123)
        :param username: optional username to connect. The default value is 'default'
        :param password: optional password to connect. The default value is empty string.
        :param pool_connections: optional number of TCP connections to pre-create when the Connection object is created.
        :param pool_maxsize: optional maximum number of TCP-connections this Connection object may make to the Clickhouse host.
        :param auth_method: 'legacy' for the Authorization header, 'x' for the X-ClickHouse-User
        :param use_own_session: True for creating a session for each Connection object, or False to reuse existing
        requests Sessions
        :param secure: whether to use TLS when connecting to the server. It is True by default, if the port is 8443,
        otherwise it is False by default (you can override the setting by passing the parameter explicitely)
        :param server_cert: if using TLS, you can pass here the file to CA Bundle or the server self-signed
        certificate. This parameter will be sent to the parameter "verify" of requests.
        :return: the Connection object
        """
        tmp = host.split(':')
        self.host = tmp[0]
        self.port = port
        if self.port is None:
            if len(tmp) > 1:
                self.port = int(tmp[-1])
            else:
                self.port = 8123
        if secure is None:
            secure = self.port == 8443
        self.secure = 's' if secure else ''
        self.server_cert = server_cert
        if isinstance(self.server_cert, str):
            self.server_cert = os.path.expanduser(self.server_cert)
        self.username = username
        self.password = password
        self.state = 'closed'
        self.timeout = timeout
        if auth_method is None and (username != 'default' or password != ''):
            auth_method = 'x'
        self.auth_method = auth_method
        self.clickhouse_settings_encoded = ''
        if len(clickhouse_settings) > 0:
            self.clickhouse_settings_encoded = '&' + '&'.join(['%s=%s' % pair for pair in list(clickhouse_settings.items())])

        self.pool_connections = pool_connections
        self.pool_maxsize = pool_maxsize

        if use_own_session:
            self.session = Connection._newsession(self.pool_connections, self.pool_maxsize)
        else:
            if Connection.Session is None or pool_connections != Connection.Pool_connections or pool_maxsize != Connection.Pool_maxsize:
                Connection.reopensession(pool_connections, pool_maxsize)
            self.session = None

    @staticmethod
    def _newsession(pool_connections=1, pool_maxsize=10):
        session = requests.Session()
        session.mount('http://', HTTPAdapter(pool_connections=pool_connections, pool_maxsize=pool_maxsize, max_retries=3))
        session.mount('https://', HTTPAdapter(pool_connections=pool_connections, pool_maxsize=pool_maxsize, max_retries=3))
        return session

    @staticmethod
    def reopensession(pool_connections=1, pool_maxsize=10):
        if Connection.Session is not None:
            Connection.Session.close()
        Connection.Session = Connection._newsession(pool_connections, pool_maxsize)
        Connection.Pool_connections = pool_connections
        Connection.Pool_maxsize = pool_maxsize

    def _call(self, query = None, payload = None):
        """
        Private method, use Cursor to make calls to Clickhouse.
        """
        try:
            credentials = self.username + ':' + self.password
            if self.auth_method == 'legacy':
                header = {'Authorization': 'Basic %s' % (base64.b64encode(credentials.encode('ISO-8859-1')),)}
            elif self.auth_method == 'x':
                header = {'X-ClickHouse-User': self.username, 'X-ClickHouse-Key': self.password}
            else:
                header = {}

            session = self.session
            if session is None:
                session = Connection.Session

            if query is None:
                return session.get('http%s://%s:%s' % (self.secure, self.host, self.port), timeout=self.timeout,
                                   headers=header, verify=self.server_cert)

            if payload is None:
                url = 'http%s://%s:%s?%s' % \
                                    (
                                        self.secure,
                                        self.host,
                                        str(self.port),
                                        self.clickhouse_settings_encoded
                                    )
                if isinstance(query, str):
                    query = query.encode('utf8')
                r = session.post(url, query, timeout=self.timeout, headers=header, verify=self.server_cert)
            else:
                url = 'http%s://%s:%s?%s' % \
                                    (
                                        self.secure,
                                        self.host,
                                        str(self.port),
                                        self.clickhouse_settings_encoded
                                    )
                if not payload.endswith('\n'):
                    payload = payload + '\n'
                if isinstance(payload, str):
                    payload = payload.encode('utf8')
                payload = query.encode('utf-8') + '\n'.encode() + payload  # on python 3, all parts must be encoded (no implicit conversion)
                r = session.post(url, payload, timeout=self.timeout, headers=header, verify=self.server_cert)
            if not r.ok:
                raise Exception('Query %s raised error %s' % (query, r.content))
            return r
        except Exception as e:
            self.close()
            try:
                if 'BadStatusLine' in str(e):  # e.g. ConnectionError has no attr. message
                    if self.session is not None:
                        self.session.close()
                        self.session = Connection._newsession(self.pool_connections, self.pool_maxsize)
                    else:
                        Connection.reopensession()
            except:
                pass
            logging.exception('When executing query %s' % query)
            raise

    def open(self):
        """
        If connection is not yet opened, checks whether Clickhouse is responding and sets the state to 'opened'
        """
        if self.state != 'opened':
            result = self._call()
            if result.content != b'Ok.\n':  # is this ok, or should we use .encode() on LHS?
                self.state = 'failed'
                raise Exception('Clickhouse not responding')
            self.state = 'opened'

    def close(self):
        """
        Closes the TCP connection pool and sets the state to 'closed'. Note that you cannot reuse this Connection object
         after calling this method.
        """
        self.state = 'closed'
        if self.session is not None:
            self.session.close()
        else:
            Connection.Session.close()


    def cursor(self):
        """
        :return: a Cursor
        """
        self.open()
        return Cursor([self])




import datetime as dt
import logging
import time
import re
import ujson

from pyclickhouse.FilterableCache import FilterableCache
from pyclickhouse.formatter import TabSeparatedWithNamesAndTypesFormatter, NestingLevelTooHigh


class Cursor(object):
    """
    Due to special design of Clickhouse, this Cursor object has a little different set of methods compared to
    typical Python database drivers.

    You can try to use it with normal pattern, like calling "execute" method first and then calling "fetchall"
    or "fetchone" afterwards. This pattern is fragile and not recommended, because the Cursor has to handle
    selects and other operations differently.

    Preferred usage pattern: call
        "select" for selects,
        "bulkinsert" for inserting many rows at once,
        "ddl" for any other statemets that don't deliver result,
        "insert" for inserting a single row (not recommended by Clickhouse)

    When calling "select", you can only use FORMAT TabSeparatedWithNamesAndTypes in your query, or omit it, in
    which case it will be added to the query automatically.

    After calling "select", you can call "fetchone" or "fetchall" to retrieve results, which will come in form
    of dictionaries.

    You can pass parameters to the queries, by marking their places in the query using %s, for example
    cursor.select('SELECT count() FROM table WHERE field=%s', 123)
    """

    def __init__(self, connections):
        """
        Create new Cursor object.
        """
        self.connections = connections
        self.connection_index = 0
        self.failed_connections = []
        self.lastresult = None
        self.lastparsedresult = None
        self.formatter = TabSeparatedWithNamesAndTypesFormatter()
        self.rowindex = -1
        self.cache = FilterableCache()


    @staticmethod
    def _escapeparameter(param):
        if isinstance(param, bool):
            return '1' if param else '0'
        if isinstance(param, int) or isinstance(param, float):
            return param
        if isinstance(param, dt.datetime):
            return "'%s'" % (str(param.replace(microsecond=0)))
        return "'%s'" % (str(param).replace("'", "\\'"))

    def execute(self, query, *args):
        """
        If possible, use one of "select", "ddl", "bulkinsert" or "insert" methods instead.
        """
        if 'select' in query.lower():
            self.select(query, *args)
        else:
            self.insert(query, *args)

    def select(self, query, *args):
        """
        Execute a select query.

        You can only use FORMAT TabSeparatedWithNamesAndTypes in your query, or omit it, in
    which case it will be added to the query automatically.

    After calling "select", you can call "fetchone" or "fetchall" to retrieve results, which will come in form
    of dictionaries.

    You can pass parameters to the queries, by marking their places in the query using %s, for example
    cursor.select('SELECT count() FROM table WHERE field=%s', 123)
        """
        if re.match(r'^.+?\s+format\s+\w+$', query.lower()) is None:
            query += ' FORMAT TabSeparatedWithNamesAndTypes'
            self.executewithpayload(query, None, True, *args)
        else:
            self.executewithpayload(query, None, False, *args)

    def select_as_dataframe(self, query, *args):
        """
        Execute a select query and parse results into a pandas dataframe

        You can only use FORMAT TabSeparatedWithNamesAndTypes in your query, or omit it, in
    which case it will be added to the query automatically.

    You can pass parameters to the queries, by marking their places in the query using %s, for example
    cursor.select_as_dataframe('SELECT count() FROM table WHERE field=%s', 123)
        """
        if re.match(r'^.+?\s+format\s+\w+$', query.lower()) is None:
            query += ' FORMAT TabSeparatedWithNamesAndTypes'
        self.executewithpayload(query, None, False, *args)
        result = self.formatter.unformat_as_dataframe(self.lastresult.content)
        return result

    def insert(self, query, *args):
        """
        Execute an insert query with data packed inside of the query parameter. Note that using "bulkinsert" can
        be more comfortable if your data is a list of dict or list of objects.
        """
        self.executewithpayload(query, None, False, *args)

    def ddl(self, query, *args):
        """
        Execute a DDL statement or other query, which doesn't return a result. Note that this statement will be
        commited automatically if succcessful.
        """
        self.executewithpayload(query, None, False, *args)

    def bulkinsert(self, table, values, fields=None, types=None):
        """
        Insert a bunch of data at once.

        :param table: Target table for inserting data, which can be optionally prepended with a database name.
        :param values: list of dictionaries or list of python objects to insert. Each key of dictionaries and
        every object property will be inserted, if fields parameter is not passed. You cannot mix dictionaries
        and objects in the values list.
        :param fields: optional list of fields to insert. Fields correspond to keys of dictionaries or properties of
        objects passed in the values parameter. If some dictionary doesn't have that key, a None value will be assumed
        :param types: optional list of strings representing Clickhouse types of corresponding fields, to ensure proper
        escaping. If omitted, the types will be inferred automatically from the first element of the values list.
        """
        fields, types, payload = self.formatter.format(values, fields, types)
        if len(payload) < 2000000000:
            self.executewithpayload('INSERT INTO %s (%s) FORMAT TabSeparatedWithNamesAndTypes' %
                                    (table, ','.join(fields)), payload, False)
        else:
            batch = int(2000000000.0 / len(payload) * len(values))
            if batch < 1:
                raise Exception("Payload of the values is larger than 2Gb, Clickhouse won't probably accept that")
            for i in range(0, len(values), batch):
                self.bulkinsert(table, values[i:i + batch], fields, types)

    def _callroundrobin(self, query, payload):
        if len(self.connections) == 1 and len(self.failed_connections) == 0:
            return self.connections[0]._call(query, payload)

        for tries in range(10):
            try:
                r = self.connections[self.connection_index]._call(query, payload)
                return r
            except:
                self.failed_connections.append(self.connections[self.connection_index])
                self.connections.remove(self.connections[self.connection_index])
                if len(self.connections) == 0:
                    self.connections = self.failed_connections
                    self.failed_connections = []
                if tries == 9:
                    raise
            finally:
                self.connection_index += 1
                if self.connection_index >= len(self.connections):
                    self.connection_index = 0

    def executewithpayload(self, query, payload, parseresult, *args):
        """
        Private method.
        """
        if args is not None and len(args) > 0:
            query = query % tuple([Cursor._escapeparameter(x) for x in args])
        self.lastresult = self._callroundrobin(query, payload)
        if parseresult and self.lastresult is not None:
            self.lastparsedresult = self.formatter.unformat(self.lastresult.content)
            self.lastresult = None  # hint GC to free memory
        else:
            self.lastparsedresult = None
        self.rowindex = -1

    def fetchone(self):
        """
        Fetch one next result row after a select query and return it as a dictionary, or None if there is no more rows.
        """
        if self.lastparsedresult is None:
            return self.lastresult.content
        if self.rowindex >= len(self.lastparsedresult) - 1:
            return None
        self.rowindex += 1
        return self.lastparsedresult[self.rowindex]

    def fetchall(self):
        """
        Fetch all resulting rows of a select query as a list of dictionaries.
        """
        return self.lastparsedresult

    def cached_select(self, query, filter):
        """
        At the first call, execute the query and store its result into a cache, organizing it in a dictionary in the way
        that rows can be retrieved efficiently, in the case the same fields are used in the filter.

        Return rows according to the filter from the cache.
        :param query: query to get and cache the values from clickhouse
        :param filter: a dictionary with keys corresponding to fields. As a value, either a scalar can be passed, or
        tuple or list, or else a slice can be passed. When scalar is passed, only rows with exact match will be
        returned. If tuple or list is passed, rows matching any of the passed values will be returned (OR principle).
        If a slice is passed, it must be either slice of int or of date. In both cases, a range of ints or dates will
        be created and rows matching the range will be returned.
        :return: The same as fetchall, a list of dictionaries
        """
        keys = sorted(filter.keys())
        tag = query + ''.join(keys)

        if not self.cache.has_dataset(tag):
            self.select(query)
            self.cache.add_dataset(tag, keys, self.fetchall())

        return self.cache.select(tag, filter)

    def get_schema(self, table):
        table = table.split('.')
        if len(table) > 2:
            raise Exception('%s is an invalid table name' % table)
        elif len(table) == 2:
            database = table[0]
            tablename = table[1]
        else:
            database = 'default'
            tablename = table[0]

        self.select('select name, type from system.columns where database=%s and table=%s', database, tablename)
        result = self.fetchall()
        return ([x['name'] for x in result], [x['type'] for x in result])

    @staticmethod
    def _flatten_array(arr, prefix='', path=[]):
        result = {}
        mapping = {}

        try:
            for i, element in enumerate(arr):
                if element is None or (hasattr(element, '__len__') and len(element) == 0):
                    continue
                if hasattr(element, 'items'):
                    r, m = Cursor._flatten_dict(element, prefix, path, allow_arrays=False)
                    for k, v in r.items():
                        if k not in result:
                            result[k] = [None] * len(arr)
                        result[k][i] = v
                    mapping.update(m)
                elif hasattr(element, '__iter__') and not isinstance(element, str):
                    raise NestingLevelTooHigh()
                else:
                    if prefix not in result:
                        result[prefix] = [None] * len(arr)
                    result[prefix][i] = element
                    mapping[prefix] = '&'.join(['%s=%s' % x for x in path])
        except NestingLevelTooHigh:
            result[prefix + '_json'] = ujson.dumps(arr)
            mapping[prefix + '_json'] = '&'.join(['%s=%s' % x for x in path[:-1] + [(path[-1][0], 'json')]])

        return result, mapping

    @staticmethod
    def _flatten_dict(doc, prefix='', path=[], allow_arrays=True):
        result = {}
        mapping = {}

        if prefix != '':
            prefix += '_'

        for k, v in doc.items():
            if v is None or (hasattr(v, '__len__') and len(v) == 0):
                continue
            if hasattr(v, 'items'):
                r, m = Cursor._flatten_dict(v, prefix + k, path + [(k, 'dict')])
                result.update(r)
                mapping.update(m)
            elif hasattr(v, '__iter__') and not isinstance(v, str):
                if allow_arrays:
                    r, m = Cursor._flatten_array(v, prefix + k, path + [(k, 'array')])
                    result.update(r)
                    mapping.update(m)
                else:
                    raise NestingLevelTooHigh()
            else:
                result[prefix + k] = v
                mapping[prefix + k] = '&'.join(['%s=%s' % x for x in path + [(k, 'scalar')]])

        return result, mapping

    def _ensure_schema(self, table, fields, types, commentmap=None, usebuffertable=None):
        tries = 0
        message = ''
        dropped = False
        while tries < 5:
            try:
                table_fields, table_types = self.get_schema(table)
                table_schema = dict(zip(table_fields, table_types))
                if usebuffertable is not None:
                    buffer_fields, buffer_types = self.get_schema(table + '_Buffer')
                    buffer_schema = dict(zip(buffer_fields, buffer_types))
                ddled = False
                new_types = []
                for doc_field, doc_type in zip(fields, types):
                    if usebuffertable is not None and \
                            (doc_field not in buffer_schema or buffer_schema[doc_field] != doc_type) and \
                            not dropped:
                        self.ddl('drop table if exists %s_Buffer' % table)
                        dropped = True
                        # it will be re-created in the store_documents

                    if doc_field not in table_schema:
                        logging.info('Extending %s with %s %s' % (table, doc_field, doc_type))
                        self.ddl('alter table %s add column %s %s' % (table, doc_field, doc_type))
                        if commentmap and doc_field in commentmap:
                            self.ddl(
                                "alter table %s comment column %s '%s'" % (table, doc_field, commentmap[doc_field]))
                        ddled = True
                        new_types.append(doc_type)
                    elif doc_field in table_schema and table_schema[doc_field] != doc_type:
                        new_type = self.formatter.generalize_type(table_schema[doc_field], doc_type)
                        if new_type != table_schema[doc_field]:
                            logging.info('Modifying %s with %s %s' % (table, doc_field, new_type))
                            self.ddl('alter table %s modify column %s %s' % (table, doc_field, new_type))
                            if commentmap and doc_field in commentmap:
                                self.ddl(
                                    "alter table %s comment column %s '%s'" % (table, doc_field, commentmap[doc_field]))
                            ddled = True
                        new_types.append(new_type)
                    else:
                        new_types.append(doc_type)

                if ddled:
                    self.ddl('optimize table %s' % table)

                return fields, new_types
            except Exception as e:
                tries += 1
                message = e.message if hasattr(e, 'message') else str(e)

        raise Exception('Cannot ensure target schema in %s, %s' % (table, message))

    def store_documents(self, table, documents,
                        nullablelambda=lambda fieldname: False,
                        usebuffertable=None,
                        extendtable=True):
        """Store dictionaries or objects into table, optionally extending the table schema if needed. If the type of 
        some value in the documents contradicts with the existing column type in clickhouse, it either will be 
        converted to String to accomodate all possible values, if extendtable is True, or the value of an 
        incompatible type will be omitted. If usebuffertable is passed (a string with the Buffer engine 
        definition), a new buffer table will be created if not exists, and used if possible. The buffer table will 
        be dropped and recreated, if new columns have to be added to the underlying table."""
        fields, flattened, types = self.prepare_document_table(table, documents, nullablelambda, usebuffertable,
                                                               extendtable)

        if usebuffertable is not None:
            self.ddl("""
            CREATE TABLE IF NOT EXISTS %s_Buffer AS %s 
            ENGINE = %s""" % (table, table, usebuffertable))

        tries = 0
        while tries < 5:
            try:
                self.bulkinsert(table if usebuffertable is None else table + '_Buffer', flattened, fields, types)
                return
            except Exception as e:
                if (hasattr(e, 'message') and 'bad version' in e.message) or 'bad version' in str(
                        e):  # can happen if we're inserting data while some other process is changing the table
                    tries += 1
                else:
                    raise

    def store_only_changed_documents(self, table, documents, primary_keys, datetimefield, ignore_fields=None,
                                     where='1=1', nullablelambda=lambda fieldname: False, usebuffertable=None):
        """
        Compares "documents" in the "table" with the latest data retrieved from the table, using grouping by the
        "primary_keys" (list of field names) and getting argMax values sorted by "datetimefield". Compares the 
        existing data with the data passed
        in "documents" ignoring the "datetimefield" as well as "ignore_fields", and inserts a new record only if some
        fields have changed. Returns the number of really inserted rows.

        Use this method only for really small tables.
        """

        table_fields, documents, table_types = self.prepare_document_table(table, documents, nullablelambda,
                                                                           usebuffertable=usebuffertable)

        if ignore_fields is None:
            ignore_fields = list()

        ignore_fields.append(datetimefield)
        ignore_fields.extend(primary_keys)

        self.select("""
        select %s, %s
        from %s
        where %s
        group by %s
        """ % (
            ','.join(primary_keys),
            ','.join(
                ['argMax(%s,%s) as %s' % (x, datetimefield, x) for x in table_fields if x not in ignore_fields]),
            table,
            where,
            ','.join(primary_keys)
        ))

        existing = dict()
        for row in self.fetchall():
            pk = tuple([row[x] for x in primary_keys])
            existing[pk] = row

        changed_documents = list()
        for doc in documents:
            pk = tuple([doc[x] for x in primary_keys])

            if pk not in existing:
                changed_documents.append(doc)
                continue

            row = existing[pk]
            for field in table_fields:
                if field in ignore_fields:
                    continue
                if field not in doc:
                    continue
                if field not in row or row[field] != doc[field]:
                    logging.info('Document with primary key %s has a change in field %s, old value %s, new value %s' % (
                        pk, field, row[field], doc[field]
                    ))
                    changed_documents.append(doc)
                    break

        if len(changed_documents) > 0:
            self.bulkinsert(table if usebuffertable is None else table + '_Buffer', changed_documents, table_fields,
                            table_types)

        return len(changed_documents)

    def _generalize_document_types(self, flattened, nullablelambda=lambda fieldname: False):
        doc_schema = {}
        for doc in flattened:
            doc_fields, doc_types = self.formatter.get_schema(doc, nullablelambda)
            for f, t in zip(doc_fields, doc_types):
                if f not in doc_schema:
                    doc_schema[f] = t
                elif doc_schema[f] != t:
                    doc_schema[f] = self.formatter.generalize_type(doc_schema[f], t)
        fields = doc_schema.keys()
        types = [doc_schema[f] for f in fields]
        return fields, types

    def _flatten_documents(self, documents):
        commentmap = {}
        if self.formatter.enable_map_datatype:
            return commentmap, documents
        flattened = []
        for doc in documents:
            f, m = Cursor._flatten_dict(doc)
            commentmap.update(m)
            flattened.append(f)
        return commentmap, flattened

    def would_change_schema(self, table, documents, nullablelambda=lambda fieldname: False):
        """Whether a subsequent call to prepare_document_table or store_documents would need
        to change table schema to accomondate the documents."""

        commentmap, flattened = self._flatten_documents(documents)
        fields, types = self._generalize_document_types(flattened, nullablelambda)

        table_fields, table_types = self.get_schema(table)
        table_schema = dict(zip(table_fields, table_types))
        for doc_field, doc_type in zip(fields, types):
            if doc_field not in table_schema or table_schema[doc_field] != doc_type:
                return True

        return False

    def prepare_document_table(self, table, documents, 
                               nullablelambda=lambda fieldname: False, 
                               usebuffertable=None,
                               extendtable=True):
        commentmap, flattened = self._flatten_documents(documents)
        if extendtable:
            fields, types = self._generalize_document_types(flattened, nullablelambda)
            fields, types = self._ensure_schema(table, fields, types, commentmap, usebuffertable)
        else:
            fields, types = self.get_schema(table)
            table_schema = dict(zip(fields, types))
            for doc in flattened:
                doc_fields, doc_types = self.formatter.get_schema(doc, nullablelambda)
                for f, t in list(zip(doc_fields, doc_types)):
                    if f not in table_schema or not self.formatter.is_compatible_type(table_schema[f], t):
                        del doc[f]

        return fields, flattened, types

    @staticmethod
    def _set_on_path(target, path, val):
        if len(path) == 0:
            raise Exception('Unexpected logical condition with path %s' % path)
        part_key, part_type = path[0].split('=')
        if part_type == 'dict':
            if part_key not in target:
                target[part_key] = {}
            Cursor._set_on_path(target[part_key], path[1:], val)
        elif part_type == 'array':
            if len(path) == 1:  # last one, means scalars
                target[part_key] = val
            else:
                if part_key not in target:
                    target[part_key] = [None] * len(val)
                for i, v in enumerate(val):
                    if i >= len(target[part_key]):
                        raise Exception('Arrays of unequal size, %s' % path)
                    if target[part_key][i] is None:
                        target[part_key][i] = dict()
                    Cursor._set_on_path(target[part_key][i], path[1:], v)
        elif part_type == 'json' and val is not None:
            target[part_key] = ujson.loads(val)
        elif val is not None:
            target[part_key] = val

    @staticmethod
    def _unflatten_dict(val, mapping):
        unflattened = {}
        for k, v in val.items():
            if k in mapping:
                path = mapping[k].split('&')
                Cursor._set_on_path(unflattened, path, v)
            else:
                unflattened[k] = v
        return unflattened

    def retrieve_documents(self, query, table_names=[]):
        self.select(query)
        rows = self.fetchall()

        if len(rows) == 0:
            return []

        self.select("""
        select name, any(comment) _comment, uniq(comment) un
        from system.columns
        where name in (%s) and length(comment) > 0 %s 
        group by name
        """ % (
            ','.join(["'%s'" % x for x in rows[0].keys()]),
            ('and table in (%s)' % ','.join(["'%s'" % x for x in table_names])) if len(table_names) > 0 else ''
        ))
        mapping = {}
        for map in self.fetchall():
            if map['un'] > 1:
                raise Exception('Cannot find a unique mapping for %s' % map['name'])
            mapping[map['name']] = map['_comment']

        return [Cursor._unflatten_dict(row, mapping) for row in rows]

    def change_and_duplicate(self, table, where, modifiers):
        """
        Select all columns of rows found by where, modify them using the SQL expressions passed in the
        modifiers dict (key: field name, value: expression), and insert them back in the same table.

        For example: change_and_duplicate('default.Items', 'id=234', {'is_deleted': '1', 'index': 'index+1'})
        """
        fields, _ = self.get_schema(table)
        sel_expr = ', '.join([modifiers[x] + ' as ' + x if x in modifiers else "`"+x+"`" for x in fields])
        self.insert("""
        insert into %s
        select %s
        from (
            select *
            from %s
            where %s
        )
        """ % (table, sel_expr, table, where))



# coding=utf-8
import unittest
import pyclickhouse
import datetime as dt
from pyclickhouse.formatter import TabSeparatedWithNamesAndTypesFormatter
import json

class TestNewUnitTests(unittest.TestCase):
    """Test compatibility of insert operations with Unicode text"""

    def setUp(self):
        self.conn = pyclickhouse.Connection('localhost:8124')
        self.cursor=self.conn.cursor()

    def test_array_serialization(self):
        self.cursor.select("select ['abc','def'] as t")
        result = self.cursor.fetchone()['t']
        assert result[0] == 'abc'
        assert result[1] == 'def'

    def test_unformat_of_commas(self):
        formatter = TabSeparatedWithNamesAndTypesFormatter()
        formatter.unformatfield("['abc',,'def']", 'Array(String)')  # boom

    def test_store_doc(self):
        doc = {'id': 3, 'historydate': dt.date(2019,6,7), 'Offer': {'price': 5, 'count': 1}, 'Images': [{'file': 'a', 'size': 400}, {'file': 'b', 'size': 500}]}
        self.cursor.ddl('drop table if exists docs')
        self.cursor.ddl('create table if not exists docs (historydate Date, id Int64) Engine=MergeTree(historydate, id, 8192)')
        self.cursor.store_documents('docs', [doc])
        self.cursor.select('select * from docs')
        r = self.cursor.fetchone()
        assert r['Images_file'] == ['a', 'b']
        assert r['Images_size'] == [400, 500]
        assert r['Offer_count'] == 1
        assert r['Offer_price'] == 5
        assert r['id'] == 3
        assert r['historydate'] == dt.date(2019, 6, 7)
        documents = self.cursor.retrieve_documents("select * from docs")
        assert doc == documents[0]

    def test_store_doc2(self):
        doc = {'id': 3, 'historydate': dt.date(2019,6,7), 'Offer': {'price': 5, 'count': 1}, 'Images': [{'file': 'a', 'size': 400, 'tags': ['cool','Nikon']}, {'file': 'b', 'size': 500}]}
        self.cursor.ddl('drop table if exists docs')
        self.cursor.ddl('create table if not exists docs (historydate Date, id Int64) Engine=MergeTree(historydate, id, 8192)')
        self.cursor.store_documents('docs', [doc])
        self.cursor.select('select * from docs')
        r = self.cursor.fetchone()
        assert 'Images_json' in r
        documents = self.cursor.retrieve_documents("select * from docs")
        assert doc == documents[0]

    def test_dict_flattening(self):
        doc = {'id': 3}
        bulk, map = pyclickhouse.Cursor._flatten_dict(doc)
        assert bulk == doc
        assert map == {'id':'id=scalar'}

        doc = {'id': 3, 'sub': {'dict': True}}
        bulk, map = pyclickhouse.Cursor._flatten_dict(doc)
        assert bulk == {'id': 3, 'sub_dict': True}
        assert map == {'id':'id=scalar', 'sub_dict': 'sub=dict&dict=scalar'}

        doc = {'id': 3, 'sub': {'dict': True, 'sub_sub': {'a': 1}}}
        bulk, map = pyclickhouse.Cursor._flatten_dict(doc)
        assert bulk == {'id': 3, 'sub_dict': True, 'sub_sub_sub_a': 1}
        assert map == {'id':'id=scalar', 'sub_dict': 'sub=dict&dict=scalar', 'sub_sub_sub_a': 'sub=dict&sub_sub=dict&a=scalar'}

        doc = {'id': 3, 'sub': ['array', 'abc']}
        bulk, map = pyclickhouse.Cursor._flatten_dict(doc)
        assert bulk == doc
        assert map == {'id':'id=scalar', 'sub': 'sub=array'}

        doc = {'id': 3, 'sub': [{'dict': 'in_array'}, {'dict': 'in_array_also', 'otherkey': True}]}
        bulk, map = pyclickhouse.Cursor._flatten_dict(doc)
        assert bulk == {'id': 3, 'sub_dict': ['in_array', 'in_array_also'], 'sub_otherkey': [None, True]}
        assert map == {'id':'id=scalar', 'sub_dict': 'sub=array&dict=scalar', 'sub_otherkey': 'sub=array&otherkey=scalar'}

        doc = {'id': 3, 'sub': {'array': ['in_dict', 'second']}}
        bulk, map = pyclickhouse.Cursor._flatten_dict(doc)
        assert bulk == {'id': 3, 'sub_array': ['in_dict', 'second']}
        assert map == {'id':'id=scalar', 'sub_array': 'sub=dict&array=array'}

        doc = {'id': 3, 'sub': [{'dict': 'in_array', 'needs': ['json', 'too_much_nesting']}]}
        bulk, map = pyclickhouse.Cursor._flatten_dict(doc)
        assert bulk['id'] == 3
        sub = json.loads(bulk['sub_json'])[0]
        assert sub['dict'] == 'in_array'
        assert sub['needs'] == ['json','too_much_nesting']
        assert map == {'id':'id=scalar', 'sub_json': 'sub=json'}


    def test_dict_unflattening(self):
        doc = {'id': 3}
        bulk = pyclickhouse.Cursor._unflatten_dict(doc,{'id':'id=scalar'})
        assert bulk == doc

        doc = {'id': 3, 'sub_dict': True}
        bulk = pyclickhouse.Cursor._unflatten_dict(doc, {'id':'id=scalar', 'sub_dict': 'sub=dict&dict=scalar'})
        assert bulk == {'id': 3, 'sub': {'dict': True}}

        doc = {'id': 3, 'sub_dict': True, 'sub_sub_sub_a': 1}
        bulk = pyclickhouse.Cursor._unflatten_dict(doc, {'id':'id=scalar', 'sub_dict': 'sub=dict&dict=scalar', 'sub_sub_sub_a': 'sub=dict&sub_sub=dict&a=scalar'})
        assert bulk == {'id': 3, 'sub': {'dict': True, 'sub_sub': {'a': 1}}}

        doc = {'id': 3, 'sub': ['array', 'abc']}
        bulk = pyclickhouse.Cursor._unflatten_dict(doc, {'id':'id=scalar', 'sub': 'sub=array'})
        assert bulk == doc

        doc = {'id': 3, 'sub_dict': ['in_array', 'in_array_also'], 'sub_otherkey': [None, True]}
        bulk = pyclickhouse.Cursor._unflatten_dict(doc, {'id':'id=scalar', 'sub_dict': 'sub=array&dict=scalar', 'sub_otherkey': 'sub=array&otherkey=scalar'})
        assert bulk == {'id': 3, 'sub': [{'dict': 'in_array'}, {'dict': 'in_array_also', 'otherkey': True}]}

        doc = {'id': 3, 'sub_array': ['in_dict', 'second']}
        bulk = pyclickhouse.Cursor._unflatten_dict(doc, {'id':'id=scalar', 'sub_array': 'sub=dict&array=array'})
        assert bulk == {'id': 3, 'sub': {'array': ['in_dict', 'second']}}

        doc = {'id': 3, 'sub_json': '[{"needs":["json","too_much_nesting"],"dict":"in_array"}]'}
        bulk = pyclickhouse.Cursor._unflatten_dict(doc, {'id':'id=scalar', 'sub_json': 'sub=json'})
        assert bulk == {'id': 3, 'sub': [{'dict': 'in_array', 'needs': ['json', 'too_much_nesting']}]}



    def test_type_generalization(self):
        types = ['Int8', 'Int16', 'Int32', 'Int64', 'Float32', 'Float64', 'Date', 'DateTime']
        results = ['Int8',
        'Int16',
        'Int32',
        'Int64',
        'Float32',
        'Float64',
        'String',
        'String',
        'Int16',
        'Int16',
        'Int32',
        'Int64',
        'Float32',
        'Float64',
        'String',
        'String',
        'Int32',
        'Int32',
        'Int32',
        'Int64',
        'Float32',
        'Float64',
        'String',
        'String',
        'Int64',
        'Int64',
        'Int64',
        'Int64',
        'Float32',
        'Float64',
        'String',
        'String',
        'Float32',
        'Float32',
        'Float32',
        'Float32',
        'Float32',
        'Float64',
        'String',
        'String',
        'Float64',
        'Float64',
        'Float64',
        'Float64',
        'Float64',
        'Float64',
        'String',
        'String',
        'String',
        'String',
        'String',
        'String',
        'String',
        'String',
        'Date',
        'DateTime',
        'String',
        'String',
        'String',
        'String',
        'String',
        'String',
        'DateTime',
        'DateTime']
        i = 0
        for left in types:
            for right in types:
                result = self.cursor.formatter.generalize_type(left, right)
                assert result == results[i]

                result = self.cursor.formatter.generalize_type('Array(%s)' % left, 'Array(%s)' % right)
                assert result == 'Array(%s)' % results[i]

                i += 1

    def test_array_generalization(self):
        result = self.cursor.formatter.clickhousetypefrompython([1, 2], 'test')
        assert result == 'Array(Int64)'

        result = self.cursor.formatter.clickhousetypefrompython([0.1, 0.2], 'test')
        assert result == 'Array(Float64)'

        result = self.cursor.formatter.clickhousetypefrompython([1, 0.2], 'test')
        assert result == 'Array(Float64)'

        result = self.cursor.formatter.clickhousetypefrompython([1, False], 'test')
        assert result == 'Array(String)'


    def test_nullable(self):
        result = self.cursor.formatter.generalize_type('Int64', 'Nullable(Int64)')
        assert result == 'Nullable(Int64)'
        result = self.cursor.formatter.generalize_type('Float64', 'Nullable(Int64)')
        assert result == 'Nullable(Float64)'

    def test_noextend(self):
        self.cursor.ddl('drop table if exists TestNoExtend ')
        self.cursor.ddl('create table TestNoExtend(id String, historydate Date) Engine=MergeTree(historydate, id, 8192)')

        doc = {'id':'first', 'historydate':dt.date.today(), 'extra':5}
        self.cursor.store_documents('TestNoExtend', [doc])
        fields, types = self.cursor.get_schema('TestNoExtend')
        assert 'extra' in fields

        doc = {'id':'second', 'historydate':dt.date.today(), 'noextra':42}
        self.cursor.store_documents('TestNoExtend', [doc], extendtable=False)
        fields, types = self.cursor.get_schema('TestNoExtend')
        assert 'noextra' not in fields

        self.cursor.ddl('drop table if exists TestNoExtend ')

    def test_nested_arrays(self):
        self.cursor.ddl('drop table if exists NestedA ')
        self.cursor.ddl('create table NestedA (a Array(Int64), b Array(String), c Array(DateTime)) Engine=Log')
        self.cursor.insert("insert into NestedA values ([1,2,3], ['a', 'b','c'], ['2023-01-01 00:00:00'])")
        docs = [{'a': [10,20,30],
                 'b': ['strinh'],
                 'c': [dt.datetime.now(), dt.datetime.now()+dt.timedelta(seconds=3)]}]
        self.cursor.bulkinsert('NestedA', docs)
        self.cursor.select("""
        select groupArray(a) aa, groupArray(b) bb, groupArray(c) cc
        from NestedA
        """)
        r = self.cursor.fetchone()
        assert r['aa'] == [[1, 2, 3], [10, 20, 30]]
        assert r['bb'] == [['a', 'b', 'c'], ['strinh']]
        assert r['cc'][0] == [dt.datetime(2023, 1, 1, 0, 0)]
        self.cursor.select('select c from NestedA limit 1')
        r = self.cursor.fetchone()
        print(r)


from setuptools import setup

REQUIRED=['requests', 'six', 'ujson', 'numpy']

setup(name='pyclickhouse3',
      version='0.9.33',
      description='Clickhouse Python driver with an API roughly resembling Python DB API 2.0 specification.',
      url='https://github.com/mfridental/PyClickhouse',
      download_url='https://github.com/mfridental/PyClickhouse/archive/0.9.33.tar.gz',
      keywords=['Clickhouse', 'Database', 'Driver'],
      classifiers=[],
      author='Maxim Fridental (Maintainer)',
      author_email='maxim@fridental.de',
      license='Apache2',
      packages=['pyclickhouse'],
      install_requires=REQUIRED,
      test_suite='test',
      zip_safe=False)




# -*- coding: utf-8 -*-
import datetime
import unittest
import datetime as dt

from pyclickhouse import Connection

class TestMap(unittest.TestCase):
    def test_formatter(self):
        from pyclickhouse.Cursor import TabSeparatedWithNamesAndTypesFormatter
        formatter = TabSeparatedWithNamesAndTypesFormatter()
        r = formatter.clickhousetypefrompython({'abc': 555, 'nope': -1}, 'values')
        assert r == 'String'
        formatter.enable_map_datatype = True
        r = formatter.clickhousetypefrompython({'abc': 555, 'nope': -1}, 'values')
        assert r == 'Map(String, Int64)'
        r = formatter.clickhousetypefrompython({'ts': datetime.datetime.now()}, 'values')
        assert r == 'Map(String, DateTime)'

    def test_db(self):
        conn = Connection('localhost:8124')
        conn.open()
        cur = conn.cursor()
        cur.formatter.enable_map_datatype = True

        cur.ddl('drop table if exists moo')
        cur.ddl('create table moo(valuemap Map(String,Float64)) Engine=Log')
        cur.insert("insert into moo values ({'prop_1': 0.2, 'prop_2': 3.14})")
        cur.bulkinsert('moo', [{'valuemap': {'prop_1': 100.0, 'prop_2': -1.0}}])
        cur.select("select sum(valuemap['prop_1']) as val from moo")
        r = cur.fetchone()
        assert r['val'] == 100.2
        cur.select("select valuemap from moo order by valuemap['prop_2']")
        r = cur.fetchall()
        assert r[0]['valuemap']['prop_1'] == 100.0
        assert r[0]['valuemap']['prop_2'] == -1.0
        assert r[1]['valuemap']['prop_1'] == 0.2
        assert r[1]['valuemap']['prop_2'] == 3.14

        cur.ddl('drop table if exists moo')
        cur.ddl('create table moo(valuemap Map(String, DateTime)) Engine=Log')
        cur.insert("insert into moo values ({'ts': '2022-01-02 00:00:34'})")
        cur.bulkinsert('moo', [{'valuemap': {'ts': dt.datetime(2023,1,1,0,0,21)}}])
        cur.select("select valuemap from moo order by valuemap['ts']")
        r = cur.fetchall()
        assert r[0]['valuemap']['ts'] == dt.datetime(2022,1,2,0,0,34)
        assert r[1]['valuemap']['ts'] == dt.datetime(2023,1,1,0,0,21)

        cur.ddl('drop table if exists moo')
        cur.ddl('create table moo(valuemap Map(String, Date)) Engine=Log')
        cur.insert("insert into moo values ({'ts': '2022-01-02'})")
        cur.bulkinsert('moo', [{'valuemap': {'ts': dt.date(2023,1,1)}}])
        cur.select("select valuemap from moo order by valuemap['ts']")
        r = cur.fetchall()
        assert r[0]['valuemap']['ts'] == dt.date(2022,1,2)
        assert r[1]['valuemap']['ts'] == dt.date(2023,1,1)



# coding=utf-8
import unittest
import pyclickhouse


class TestPy23Compatibility(unittest.TestCase):
    """Test compatibility of insert operations with Unicode text"""

    def setUp(self):
        self.conn = pyclickhouse.Connection('localhost:8124')
        self.cursor=self.conn.cursor()
        self.cursor.ddl("create database if not exists test")
        self.cursor.ddl("drop table if exists test")


    def tearDown(self):
        self.cursor.ddl("drop database if exists test")

    def test_insert(self):
        self.cursor.ddl("create table IF NOT EXISTS test.test (geoid String, objektanzahl Int64) Engine=TinyLog")
        self.cursor.insert(u"insert into test.test values ('ŋéöìð',123)")

    def test_bulkinsert(self):
        self.cursor.ddl("create table IF NOT EXISTS test.test (geoid String, objektanzahl Int64) Engine=TinyLog")
        values = [dict(geoid=u'ŋéöìð', objektanzahl=123)]
        self.cursor.bulkinsert(table='test.test', values=values)

    def test_insert_list(self):
        self.cursor.ddl("create table IF NOT EXISTS test.test_list (geoid Array(String), objektanzahl Int64) Engine=TinyLog")
        self.cursor.insert(u"insert into test.test_list values (array('ŋéöìð','ŋéöìð'),123)")

    def test_bulkinsert_list(self):
        self.cursor.ddl("create table IF NOT EXISTS test.test_list (geoid Array(String), objektanzahl Int64) Engine=TinyLog")
        values = [dict(geoid=[u'ŋéöìð',u'ŋéöìð'], objektanzahl=123)]
        self.cursor.bulkinsert(table='test.test_list', values=values)




# -*- coding: utf-8 -*-
import datetime
import unittest
import datetime as dt

from pyclickhouse import Connection

class TestSimple(unittest.TestCase):
    def test_simple(self):
        conn = Connection('localhost:8124')
        conn.open()
        cur = conn.cursor()

        cur.ddl("""
        drop table if exists simpletest
        """)

        cur.select('select count() from system.tables')
        cnt = cur.fetchall()

        print(cnt)

        cur.select('select * from system.tables ')
        print(len(cur.fetchall()))

        cur.ddl("""
        create table simpletest (
        f01 UInt8,
        f02 UInt16,
        f03 UInt32,
        f04 UInt64,
        f05 Int8,
        f06 Int16,
        f07 Int32,
        f08 Int64,
        f09 Float32,
        f10 Float64,
        f11 String,
        f12 Date,
        f13 DateTime,
        f14 Array(UInt8),
        f15 Array(UInt16),
        f16 Array(UInt32),
        f17 Array(UInt64),
        f18 Array(Int8),
        f19 Array(Int16),
        f20 Array(Int32),
        f21 Array(Int64),
        f22 Array(Float32),
        f23 Array(Float64),
        f24 Array(String),
        f25 Array(Date),
        f26 Array(DateTime),
        f27 DateTime('UTC')
        )
        engine=Memory
        """)

        cur.select('select count() from system.tables')
        assert cur.fetchone()['count()'] == cnt[0]['count()'] + 1

        cur.insert("""
        insert into simpletest values
        (
        1,2,3,4,5,6,7,8,9.99, 10.10, 'abcdeföäüß', '1975-11-10', '1975-11-10 02:23:01',
        [10,12,13], [10,12,13],[10,12,13],[10,12,13],[10,12,13],[10,12,13],[10,12,13],[10,12,13],
        [10.1,12.2,13.3],[10.1,12.2,13.3],
        ['escaping','te\tst'],
        ['1975-11-10'], ['1975-11-10 02:23:01'], '1975-11-10'
        )
        """)

        cur.select("""select count() from simpletest """)
        assert cur.fetchone()['count()'] == 1

        values = [
            {
                'f01': 100,
                'f02': 101,
                'f03': 102,
                'f04': 103,
                'f05': 104,
                'f06': 105,
                'f07': 106,
                'f08': 107,
                'f09': 109.9,
                'f10': 110.10,
                'f11': 'a string',
                'f12': datetime.datetime(1975,11,10).date(),
                'f13': datetime.datetime(1975,11,10,2,23,1),
                'f14': [101],
                'f15': [102],
                'f16': [103],
                'f17': [104],
                'f18': [105],
                'f19': [106],
                'f20': [107],
                'f21': [108],
                'f22': [109.99],
                'f23': [111.11],
                'f24': ['string in array'],
                'f25': [dt.date.today()],
                'f26': [dt.datetime.now()],
                'f27': dt.datetime.utcnow()
            }]

        #values = values * 10000

        start = dt.datetime.now()
        cur.bulkinsert('simpletest', values, fields=values[0].keys(), types=[
            'UInt8',
            'UInt16',
            'UInt32',
            'UInt64',
            'Int8',
            'Int16',
            'Int32',
            'Int64',
            'Float32',
            'Float64',
            'String',
            'Date',
            'DateTime',
            'Array(UInt8)',
            'Array(UInt16)',
            'Array(UInt32)',
            'Array(UInt64)',
            'Array(Int8)',
            'Array(Int16)',
            'Array(Int32)',
            'Array(Int64)',
            'Array(Float32)',
            'Array(Float64)',
            'Array(String)',
            'Array(Date)',
            'Array(DateTime)',
            "DateTime('UTC')"
        ])
        print(10000.0 / (dt.datetime.now() - start).total_seconds())

        cur.select("""
        select groupArray(f26[1]) arr from simpletest
        """)
        print(cur.fetchone()['arr'])

        cur.select("""
        select count() from simpletest
        """)
        cnt = cur.fetchone()['count()']

        cur.bulkinsert('simpletest', [{'f01': 100, 'f02': 101, 'f03': 102}], ['f01', 'f02', 'f03', 'f04'],
                       ['UInt8', 'UInt16', 'UInt32', 'UInt64'])

        cur.bulkinsert('simpletest', [{}], ['f14'], ['Array(UInt8)'])

        cur.bulkinsert('simpletest', [{'f11': "can't"}], ['f11'], ['String'])

        cur.bulkinsert('simpletest', [{'f24': ["can't"]}], ['f24'], ['Array(String)'])

        cur.select("""
        select count() from simpletest
        """)
        cnt2 = cur.fetchone()['count()']

        assert cnt2 == cnt + 4

        cur.select("""
        select f27 from simpletest limit 1
        """)
        val = cur.fetchone()['f27']
        assert val.tzname() == 'UTC'




# -*- coding: utf-8 -*-
import sys
import logging
import datetime as dt

class FilterableCache(object):
    def __init__(self):
        self.data = {}
        self.index = {}

    def has_dataset(self, dataset_key):
        return dataset_key in self.data

    def add_dataset(self, dataset_key, primarykeys, rows):
        self.data[dataset_key] = dict([(tuple([row[field] for field in primarykeys]), row) for row in rows])
        key_info = []
        for i, _ in enumerate(primarykeys):
            key_info.append(set([x[i] for x in self.data[dataset_key].keys()]))
        self.index[dataset_key] = key_info

    def _recursivefoo(self, result, key, level, dataset_key):
        if level >= len(key):
            if level < len(self.index[dataset_key]):
                for existing_val in self.index[dataset_key][level]:
                    subkey = list(key)
                    subkey.append(existing_val)
                    subkey = tuple(subkey)
                    self._recursivefoo(result, subkey, level + 1, dataset_key)
            else:
                if key in self.data[dataset_key]:
                    result.append(self.data[dataset_key][key])
        else:
            if isinstance(key[level], slice):
                if isinstance(key[level].start, int) and isinstance(key[level].stop, int):
                    for x in range(key[level].start, key[level].stop):
                        subkey = list(key)
                        subkey[level] = x
                        subkey = tuple(subkey)
                        self._recursivefoo(result, subkey, level+1, dataset_key)
                elif isinstance(key[level].start, dt.date) and isinstance(key[level].stop, dt.date):
                    for x in range(0, (key[level].stop - key[level].start).days):
                        subkey = list(key)
                        subkey[level] = key[level].start + dt.timedelta(days=x)
                        subkey = tuple(subkey)
                        self._recursivefoo(result, subkey, level+1, dataset_key)
                else:
                    raise Exception('Only tuples of int or datetime.date are supported in filter')
            elif isinstance(key[level], list) or isinstance(key[level], tuple):
                for x in key[level]:
                    subkey = list(key)
                    subkey[level] = x
                    subkey = tuple(subkey)
                    self._recursivefoo(result, subkey, level + 1, dataset_key)
            else:
                self._recursivefoo(result, key, level + 1, dataset_key)

    def select(self, dataset_key, filter):
        keys = sorted(filter.keys())
        filtertuple = tuple([filter[field] for field in keys])
        result = []
        self._recursivefoo(result, filtertuple, 0, dataset_key)
        return result

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)

    data = FilterableCache()
    data.add_dataset('test', ['Color','Size'], [
        {'Color': 'red', 'Size': 10, 'Price': 300},
        {'Color': 'red', 'Size': 11, 'Price': 330},
        {'Color': 'red', 'Size': 12, 'Price': 360},
        {'Color': 'green', 'Size': 10, 'Price': 300},
        {'Color': 'green', 'Size': 11, 'Price': 330},
        {'Color': 'brown', 'Size': 10, 'Price': 100}])

    print(data.select('test', {'Color': 'red', 'Size': 10}))
    print()
    print(data.select('test', {'Color': 'red', 'Size': (10, 12, 555)}))
    print()
    print(data.select('test', {'Color': 'red', 'Size': slice(10, 12)}))
    print()
    print(data.select('test', {'Color': ['red', 'green', 'not there'], 'Size': 10}))
    print()
    print(data.select('test', {'Color': 'red'}))
    print()